#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : state.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

from datetime import datetime
from enum import IntEnum, auto

from discord.ext import commands


class GameMode(IntEnum):
    DEFAULT = auto()
    RT = auto()
    DND = auto()
    AW = auto()


MODE_META = {GameMode.DEFAULT: 'Default die is 1d6',
             GameMode.RT: 'Default die is 1d100; reports DoS/DoF',
             GameMode.DND: 'Default die is 1d20',
             GameMode.AW: 'Default die is 1d6'}


MODE_DICE = {GameMode.DEFAULT: '1d6',
             GameMode.RT: '1d100',
             GameMode.DND: '1d20',
             GameMode.AW: '2d6t'}

MAX_DICE_PER_ROLL = 64



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : zmode.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import typing

from discord.ext import commands

from ..database import fetch_guild_db
from ..logging import LoggingMixin
from ..state import GameMode, MODE_META


class ModeConvert(commands.Converter):

    async def convert(self, ctx, argument):
        try:
            converted = GameMode[argument]
        except KeyError:
            raise commands.BadArgument(f'{argument} is not a valid mode.')
        return converted


class ModeCommands(commands.Cog, LoggingMixin):

    def __init__(self, bot, db):
        self.bot = bot
        self.db = db

        super().__init__()

    @commands.group(name='mode')
    @fetch_guild_db
    async def mode(self, ctx, mode: typing.Optional[ModeConvert]):
        '''
        Get and set the game mode for this server.
        '''
        if ctx.invoked_subcommand is not None:
            return

        if mode is None:
            current_mode = await ctx.guild_db.get_guild_mode()
            await ctx.send(f'**Mode:**: {GameMode(current_mode).name}\n'\
                           f'*{MODE_META[current_mode]}*')
        else:
            try:
                await ctx.guild_db.set_guild_mode(mode)
            except ValueError as e:
                await ctx.message.reply(f'{e}')
            else:
                await ctx.message.reply(f'**Set Mode:** {GameMode(mode).name}')
    
    @mode.command(name='list')
    async def list(self, ctx):
        '''
        List the available game modes for this server.
        '''
        modes = '\n'.join((f'{mode.name}: {MODE_META[mode]}' for mode in GameMode))
        await ctx.message.reply(modes)




from __future__ import absolute_import

from pyparsing import Literal, ParseFatalException
from pytest import raises
import random
import string

from dice import roll, utilities
from dice.utilities import verbose_print
from dice.elements import RandomElement, FudgeDice


def test_enable_pyparsing_packrat_parsing():
    """Test that packrat parsing was enabled"""
    import pyparsing

    assert pyparsing.ParserElement._packratEnabled is True


def test_disable_pyparsing_arity_trimming():
    """Test that pyparsing._trim_arity has been replaced"""
    import pyparsing
    import dice.utilities

    assert pyparsing._trim_arity is dice.utilities._trim_arity


def test_disable_pyparsing_arity_trimming_works():
    """Tests that arity trimming has been disabled and parse actions with
    the wrong number of arguments will raise TypeErrors"""
    for func in [lambda a: None, lambda a, b: None, lambda a, b, c, d: None]:
        element = Literal("test").setParseAction(func)
        with raises(TypeError):
            element.parseString("test")


class TestVerbosePrint(object):
    def _get_vprint(self, expr):
        raw = roll(expr, raw=True)
        evaluated = raw.evaluate_cached()
        vprint = verbose_print(raw)
        return evaluated, vprint

    def test_dice_simple(self):
        v = self._get_vprint("6d6")[1]
        assert v.startswith("roll 6d6 -> ")

    def test_dice_complex(self):
        v = self._get_vprint("6d(6d6)t")[1]
        lines = v.split("\n")
        stripped = [l.strip() for l in lines]

        assert len(lines) == 6
        assert lines[0] == "Total("
        assert stripped[1] == "Dice("
        assert stripped[2] == "6,"
        assert stripped[3].startswith("roll 6d6")
        assert stripped[4].startswith(") -> ")
        assert stripped[5].startswith(") -> ")

    def test_unevaluated(self):
        r = roll("d20", raw=True)
        v = verbose_print(r)
        assert v.startswith("roll 1d20 -> ")

    def test_single_line(self):
        v = self._get_vprint("d6x")[1]
        assert len(v.split("\n")) == 1


class TestDiceSwitch(object):
    def test_separator_map(self):
        for sep, cls in RandomElement.DICE_MAP.items():
            d = utilities.dice_switch(6, 6, sep)
            assert type(d) is cls

    def test_percentile(self):
        for sep, cls in RandomElement.DICE_MAP.items():
            d = utilities.dice_switch(6, "%", sep)
            assert d.sides == 100

    def test_fudge(self):
        assert type(utilities.dice_switch(1, "f", "d")) == FudgeDice
        assert type(utilities.dice_switch(1, "f", "u")) == FudgeDice

        with raises(ValueError):
            utilities.dice_switch(1, "f", "w")

    def test_invalid(self):
        unused = set(string.ascii_lowercase) - set(RandomElement.DICE_MAP)

        bad_params = [(1, 0, "d"), (1, 6, "dd")]

        if unused:
            unused_char = random.choice(list(unused))
            bad_params.append((1, 6, unused_char))

        for params in bad_params:
            with raises(ValueError):
                utilities.dice_switch(*params)


def test_binary_stack():
    with raises(ParseFatalException):
        roll("6d6d6")


def test_too_many_explosions():
    with raises(ParseFatalException):
        while True:
            roll("1000d1000x2")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# zardoz/zardoz.py
# Copyright (c) 2021 Camille Scott <camille.scott.w@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import json
import logging
import os
from pathlib import Path
import sys
import textwrap

from discord.ext import commands

from . import __version__, __splash__, __about__, __testing__
from .utils import default_log_file, default_database_dir, SYNTAX


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


class Bot(commands.Bot):

    async def close(self):
        log = logging.getLogger()
        log.info('DatabaseCache close()')
        await self.DB.close()
        log.info('Bot close()')
        await super().close()
        log.info('Bot closed.')


def add_parser_args(parser, token_name='ZARDOZ_TOKEN'):
    parser.add_argument(
        '--secret-token',
        help=f'The secret token for the bot, '\
             f'from the discord developer portal. '\
             f'If you have set ${token_name}, this '\
             f'option will override that.'
    )
    parser.add_argument(
        '--database-dir',
        type=lambda p: Path(p).absolute(),
        default=default_database_dir(debug=__testing__),
        help='Directory for the bot databases. Default follows the '\
             'XDG specifiction.'
    )
    parser.add_argument(
        '--log',
        type=lambda p: Path(p).absolute(),
        default=default_log_file(debug=__testing__),
        help='Path to the log file. Default follows the '\
             'XDG specifiction.'
    )


def main():
    parser = argparse.ArgumentParser(
        description=f'{__splash__}\n{__about__}',
        formatter_class=CustomFormatter
    )
    subparsers = parser.add_subparsers()

    bot = subparsers.add_parser('bot')
    add_parser_args(bot)
    bot.set_defaults(func=run_bot)

    args = parser.parse_args()
    args.func(args)


def run_bot(args):

    bot, db = build_bot(args)

    @bot.event
    async def on_ready():
        log = logging.getLogger()
        log.info(f'Ready: member of {bot.guilds}')
        log.info(f'Users: {bot.users}')

    @bot.command(name='about', help='Project info.')
    async def zabout(ctx):
        msg = f'version: {__version__}\n'\
              f'source: https://github.com/camillescott/zardoz/releases/tag/v{__version__}\n'\
              f'active installs: {len(bot.guilds)}'
        await ctx.message.reply(msg)

    @bot.command(name='syntax', help='Rolling syntax')
    async def zsyntax(ctx):
        await ctx.message.reply(SYNTAX)

    bot.run(args.secret_token)


def build_bot(args, token_name='ZARDOZ_TOKEN', prefix='z', loop=None):

    from .database import DatabaseCache
    from .logging import setup as setup_logger
    
    from .cogs.history import HistoryCommands
    from .cogs.mode import ModeCommands
    from .cogs.roll import RollCommands
    from .cogs.sample import SampleCommands
    from .cogs.vars import VarCommands

    log = setup_logger(args.log)

    TOKEN = args.secret_token
    if not TOKEN:
        try:
            TOKEN = os.environ[token_name]
        except KeyError:
            log.error(f'Must set ${token_name} or use --secret-token')
            sys.exit(1)
        else:
            log.info(f'Got secret token from ${token_name}.')
    args.secret_token = TOKEN

    # get a handle for the history database
    DB = DatabaseCache(args.database_dir)

    prefix = f'/{prefix}' if not __testing__ else f'!{prefix}'
    log.info(f'Prefix is: {prefix}')
    bot = Bot(command_prefix=prefix, loop=loop)
    bot.DB = DB

    bot.add_cog(RollCommands(bot, DB))
    bot.add_cog(VarCommands(bot, DB))
    bot.add_cog(ModeCommands(bot, DB))
    bot.add_cog(HistoryCommands(bot, DB))
    bot.add_cog(SampleCommands(bot, DB))

    return bot, DB


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : zroll.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import typing

import discord
from discord.ext import commands

from ..database import fetch_guild_db
from ..logging import LoggingMixin
from ..rolls import (RollHandler, QuietRollHandler, SekretRollHandler, RerollHandler,
                    tokenize_roll)
from ..utils import handle_http_exception


class RollCommands(commands.Cog, LoggingMixin):

    def __init__(self, bot, db):
        self.bot = bot
        self.db = db

        super().__init__()

    @commands.command(name='r', aliases=['roll'])
    @fetch_guild_db
    @handle_http_exception
    async def roll(self, ctx, *, args):
        '''
        Evaluate a dice roll and reply to the requester with the result
        in the source channel.
        '''

        try:
            roll = RollHandler(ctx, self.log, ctx.variables, args,
                               game_mode=ctx.game_mode)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.message.reply(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(roll.msg())

    @commands.command(name='q', aliases=['quiet'])
    @fetch_guild_db
    @handle_http_exception
    async def quiet_roll(self, ctx, *, args):
        '''
        Evaluate a dice roll and reply to the requester with a simplified result
        in the source channel.
        '''
        try:
            roll = QuietRollHandler(ctx, self.log, ctx.variables, args,
                                    game_mode=ctx.game_mode)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.message.reply(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(roll.msg())

    @commands.command(name='s', aliases=['secret'])
    @fetch_guild_db
    @handle_http_exception
    async def secret_roll(self, ctx, member: typing.Optional[discord.Member], *, args):
        '''
        Evaluate a dice roll and DM the result to the given member,
        '''

        if member is None:
            member = ctx.author

        try:
            roll = SekretRollHandler(ctx, self.log, ctx.variables, args,
                                     game_mode=ctx.game_mode, require_tag=True)
        except ValueError as e:
            self.log.error(f'Roll handling failed: {e}')
            await ctx.author.send(f'You fucked up your roll, {ctx.author}. {e}')
        else:
            await roll.add_to_db(ctx.guild_db)
            await member.send(roll.msg())

    @commands.command(name='rr', aliases=['reroll'])
    @fetch_guild_db
    @handle_http_exception
    async def zroll_reroll(self, ctx, member: typing.Optional[discord.Member], *, args=''):
        '''
        Re-evaluate the previous roll of the given member, or of the requesting member
        if not given, and reply with the result in the source channel.
        '''

        if member is None:
            member = ctx.author

        saved = await ctx.guild_db.get_last_user_roll(member.id)

        if saved is None:
            await ctx.message.reply(f'Ope, no roll history for {member}.')
        else:
            cmd = saved['roll']
            _, tag = tokenize_roll(args)
            if not tag:
                tag = saved['tag']
            tag = '# ' + tag

            roll = RerollHandler(ctx, self.log, ctx.variables, cmd + tag,
                                 game_mode=ctx.game_mode)

            if member == ctx.author:
                target = 'self'
            else:
                target = member.nick if member.nick else member.name

            await roll.add_to_db(ctx.guild_db)
            await ctx.message.reply(f'Reroll {roll.msg(target)}')



from __future__ import absolute_import

from dice import roll, roll_min, roll_max
from dice.command import main
from itertools import product
from pytest import raises


def test_roll():
    for single, raw in product((True, False), (True, False)):
        assert roll("6d6", single=single, raw=raw)
        assert roll_min("6d6", single=single, raw=raw)
        assert roll_max("6d6", single=single, raw=raw)


def test_main():
    main(["2d6"])


def test_main_verbose():
    main(["2d6", "--verbose"])


def test_main_min():
    main(["2d6", "--min"])


def test_main_max():
    main(["2d6", "--max"])


def test_main_max_dice():
    main(["2d6", "--max-dice", "2"])


def test_main_max_dice_err():
    with raises(SystemExit):
        main(["2d6", "--max-dice", "not_a_number"])


def test_main_error():
    with raises(SystemExit):
        main(["d0"])


def test_main_error2():
    """Test placing the error on the left"""
    with raises(SystemExit):
        main(["000000000000000000000000000000000000000001d6, d0"])



from pyparsing import ParseException, ParseFatalException


class DiceBaseException(Exception):
    @classmethod
    def from_other(cls, other):
        if isinstance(other, ParseException):
            return DiceException(*other.args)
        elif isinstance(other, ParseFatalException):
            return DiceFatalException(*other.args)
        raise NotImplementedError(
            'DiceBaseException can only wrap ParseException or ParseFatalException'
        )

    def pretty_print(self):
        string, location, description = self.args
        lines = string.split("\n")

        if len(description) < (self.col - 1):
            line = (description + " ^").rjust(self.col)
        else:
            line = "^ ".rjust(self.col + 1) + description

        lines.insert(self.lineno + 1, line)
        return "\n".join(lines)


class DiceException(DiceBaseException, ParseException):
    pass


class DiceFatalException(DiceBaseException, ParseFatalException):
    pass



"""Objects used in the evaluation of the parse tree"""

from __future__ import absolute_import, print_function, unicode_literals

import random
import operator
from pyparsing import ParseFatalException
from copy import copy

from .constants import MAX_EXPLOSIONS, MAX_ROLL_DICE, DiceExtreme
from .exceptions import DiceFatalException
from .utilities import classname, add_even_sub_odd, dice_switch

from pdb import set_trace


class Element(object):
    @classmethod
    def parse(cls, string, location, tokens):
        try:
            return cls(*tokens).set_parse_attributes(string, location, tokens)
        # The following only matters on python2 platforms, so is marked nocover
        except ArithmeticError as e:  # nocover
            exc = DiceFatalException(string, location, e.args[0])  # nocover
            exc.__cause__ = None  # nocover
            raise exc  # nocover

    def set_parse_attributes(self, string, location, tokens):
        self.string = string
        self.location = location
        self.tokens = tokens
        return self

    def fatal(self, description, location=None, offset=0, cls=DiceFatalException):
        if location is None:
            location = self.location
        return cls(self.string, location + offset, description)

    def evaluate(self, **kwargs):
        """Evaluate the current object - a no-op by default"""
        return self

    @staticmethod
    def evaluate_object(obj, cls=None, cache=False, **kwargs):
        """Evaluates elements, and coerces objects to a class if needed"""
        old_obj = obj
        if isinstance(obj, Element):
            if cache:
                obj = obj.evaluate_cached(**kwargs)
            else:
                obj = obj.evaluate(cache=cache, **kwargs)

        if cls is not None and type(obj) != cls:
            obj = cls(obj)

        for attr in ("string", "location", "tokens"):
            if hasattr(old_obj, attr):
                setattr(obj, attr, getattr(old_obj, attr))

        return obj

    def evaluate_cached(self, **kwargs):
        """Wraps evaluate(), caching results"""
        if not hasattr(self, "result"):
            self.result = self.evaluate(cache=True, **kwargs)

        return self.result


class Integer(int, Element):
    """A wrapper around the int class"""

    pass


class String(str, Element):
    """A wrapper around the str class"""

    pass


class Comparison(Element):

    def __init__(self, left, right, delta, value, operator):
        self.left = left
        self.right = right
        self.delta = delta
        self.value = value
        self.operator = operator

    def __repr__(self):
        return f'<Comparison: {self.left} {self.operator} {self.right} -> {self.value}, {self.delta}>'


class IntegerList(list, Element):
    "Augments the standard list with an __int__ operator"

    def __str__(self):
        ret = "[%s]" % ", ".join(map(str, self))
        if hasattr(self, "sum") and len(self) > 1:
            ret += " -> %i" % self
        return ret

    def copy(self):
        return type(self)(self)

    def clear(self):
        self[:] = []

    def __int__(self):
        ret = sum(self)
        self.sum = ret
        return ret


class Roll(IntegerList):
    """Represents a randomized result from a random element"""

    @classmethod
    def roll_single(cls, min_value, max_value, **kwargs):
        integer_min = cls.evaluate_object(min_value, Integer, **kwargs)
        integer_max = cls.evaluate_object(max_value, Integer, **kwargs)

        if integer_min > integer_max:
            raise ValueError(
                "Roll must have a valid range (got %s - %s, "
                "which evaluated to %i - %i). Are you trying to "
                "use a fudge roll as the sides?"
                % (min_value, max_value, integer_min, integer_max)
            )
        rnd_engine = kwargs.get("random", random)
        return rnd_engine.randint(integer_min, integer_max)

    @classmethod
    def roll(cls, orig_amount, min_value, max_value, **kwargs):
        amount = cls.evaluate_object(orig_amount, Integer, **kwargs)

        max_dice = kwargs.get("max_dice", MAX_ROLL_DICE)

        if amount > max_dice:
            raise ValueError("Too many dice! (max is %i)" % max_dice)
        elif amount < 0:
            raise ValueError("Cannot roll less than zero dice!")

        return [cls.roll_single(min_value, max_value, **kwargs) for i in range(amount)]

    def do_roll_single(self, min_value=None, max_value=None, **kwargs):
        element = self.random_element

        if self.force_extreme is DiceExtreme.EXTREME_MIN:
            return element.min_value
        elif self.force_extreme is DiceExtreme.EXTREME_MAX:
            return element.max_value

        if min_value is None:
            min_value = element.min_value

        if max_value is None:
            max_value = element.max_value
        try:
            return self.roll_single(min_value, max_value, **kwargs)
        except ValueError as e:
            exc = self.random_element.fatal(e.args[0])
            raise exc

    def do_roll(self, amount=None, min_value=None, max_value=None, **kwargs):
        element = self.random_element
        if amount is None:
            amount = element.amount
        if min_value is None:
            min_value = element.min_value
        if max_value is None:
            max_value = element.max_value

        try:
            return self.roll(amount, min_value, max_value, **kwargs)
        except ValueError as e:
            exc = self.random_element.fatal(e.args[0])
            exc.__cause__ = None
            raise exc

    def __init__(self, element, rolled=None, **kwargs):
        self.random_element = element
        self.force_extreme = kwargs.get("force_extreme")

        amount = self.evaluate_object(element.amount, Integer, **kwargs)
        min_value = self.evaluate_object(element.min_value, Integer, **kwargs)
        max_value = self.evaluate_object(element.max_value, Integer, **kwargs)

        if rolled is None:
            max_dice = kwargs.get("max_dice", MAX_ROLL_DICE)

            if amount > max_dice:
                msg = "Too many dice! (max is %i)" % max_dice
                exc = self.random_element.fatal(msg)
                exc.__cause__ = None
                raise exc
            elif amount < 0:
                msg = "Cannot roll less than zero dice!"

                if not isinstance(element.amount, int):
                    msg += " (%s evaluated to %s)" % (element.amount, amount)

                exc = self.random_element.fatal(msg)
                exc.__cause__ = None
                raise exc

            elif self.force_extreme is DiceExtreme.EXTREME_MIN:
                rolled = [min_value] * amount
            elif self.force_extreme is DiceExtreme.EXTREME_MAX:
                rolled = [max_value] * amount
            else:
                rolled = self.do_roll(**kwargs)

        super(Roll, self).__init__(rolled)

    def copy(self):
        return type(self)(
            self.random_element, rolled=self, force_extreme=self.force_extreme
        )

    def __repr__(self):
        return "{0}({1}, random_element={2!r})".format(
            classname(self), str(self), self.random_element
        )


class WildRoll(Roll):
    """Represents a roll of wild dice"""

    @classmethod
    def roll(cls, amount, min_value, max_value, **kwargs):
        amount = cls.evaluate_object(amount, Integer, **kwargs)
        min_value = cls.evaluate_object(min_value, Integer, **kwargs)
        max_value = cls.evaluate_object(max_value, Integer, **kwargs)

        if amount == 0:
            return []

        rnd_engine = kwargs.get("random", random)
        rolls = [rnd_engine.randint(min_value, max_value) for i in range(amount)]

        if min_value == max_value:
            return rolls  # Continue as if dice were normal

        while rolls[-1] == max_value:
            rolls.append(rnd_engine.randint(min_value, max_value))

        if len(rolls) == amount and rolls[-1] == min_value:  # failure
            rolls[-1] = 0
            rolls[rolls.index(max(rolls))] = 0

            if rnd_engine.randint(min_value, max_value) == min_value:
                return [0] * amount

        return rolls


class ExplodedRoll(Roll):
    """Represents an exploded roll"""

    def __init__(self, original, rolled, **kwargs):
        super(ExplodedRoll, self).__init__(original, rolled=rolled, **kwargs)


class RandomElement(Element):
    """Represents a set of elements with a random numerical result"""

    DICE_MAP = {}
    SEPARATOR = None

    @classmethod
    def register_dice(cls, new_cls):
        if not issubclass(new_cls, RandomElement):
            raise TypeError("can only register subclasses of RandomElement")
        elif not new_cls.SEPARATOR:
            raise TypeError("must specify separator")
        elif new_cls.SEPARATOR in cls.DICE_MAP:
            raise RuntimeError("Separator %s already registered" % new_cls.SEPARATOR)
        cls.DICE_MAP[new_cls.SEPARATOR] = new_cls
        return new_cls

    @classmethod
    def parse_unary(cls, string, location, tokens):
        return cls.parse(string, location, [1] + list(tokens))

    @classmethod
    def parse(cls, string, location, tokens):
        if len(tokens) > 3:
            raise ParseFatalException(
                string,
                tokens[3].location,
                (
                    "Cannot stack dice operators! Try disambiguating your "
                    "expression with parentheses."
                ),
            )

        amount, kind, dice_type = tokens
        try:
            ret = dice_switch(amount, dice_type, kind)
            return ret.set_parse_attributes(string, location, tokens)
        except ValueError as e:
            if len(e.args) > 1:
                if type(e.args[1]) is int:
                    location = tokens[e.args[1]].location
                # unused as of yet
                # elif isinstance(e.args[1], Element):
                #     location = e.args[1].location
            raise ParseFatalException(string, location, e.args[0])

    @classmethod
    def from_iterable(cls, iterable):
        return cls(*iterable)

    @classmethod
    def from_string(cls, string):
        string = string.lower()
        for k in cls.DICE_MAP:
            ss = string.split(k)
            if len(ss) != 2:
                continue
            ss = [int(x) if x.isdigit() else x for x in ss]
            return dice_switch(ss[0], ss[1], k)

    def __init__(self, amount, min_value, max_value, **kwargs):
        self.amount = amount
        self.min_value = min_value
        self.max_value = max_value

    def __neg__(self):
        new = copy(self)
        new.min_value, new.max_value = -new.max_value, -new.min_value
        return new

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.amount == other.amount
            and self.min_value == other.min_value
            and self.max_value == other.max_value
        )

    def evaluate(self, **kwargs):
        return Roll(self, **kwargs)


@RandomElement.register_dice
class Dice(RandomElement):
    """A group of dice, all with the same number of sides"""

    SEPARATOR = "d"

    def __init__(self, amount, max_value, min_value=1):
        super(Dice, self).__init__(amount, min_value, max_value)
        self.original_operands = (amount, max_value)

    @property
    def sides(self):
        return self.max_value

    def __repr__(self):
        p = "{0!r}, {1!r}".format(self.amount, self.max_value)
        if self.min_value != 1:
            p += ", {0!r}".format(self.min_value)
        return "{}({})".format(classname(self), p)

    def __str__(self):
        return "{0!s}{1}{2!s}".format(self.amount, self.SEPARATOR, self.sides)


@RandomElement.register_dice
class WildDice(Dice):
    """A group of dice with the last being explodable or a failure mode on 1."""

    SEPARATOR = "w"

    def evaluate(self, **kwargs):
        return WildRoll(self, **kwargs)


@RandomElement.register_dice
class FudgeDice(Dice):
    """A group of dice whose sides range from -x to x, including 0"""

    SEPARATOR = "u"

    def __init__(self, amount, value):
        super(FudgeDice, self).__init__(amount, value, -value)

    def __repr__(self):
        p = "{0!r}, {1!r}".format(self.amount, self.max_value)

        if self.min_value != -self.max_value:
            p += ", {0!r}".format(self.min_value)

        return "{}({})".format(classname(self), p)


class Operator(Element):
    PASS_KWARGS = ()

    def __init__(self, *operands):
        self.operands = self.original_operands = operands

    def __repr__(self):
        return "{0}({1})".format(
            classname(self), ", ".join(map(str, self.original_operands))
        )

    def __getnewargs__(self):
        return self.original_operands

    def preprocess_operands(self, *operands, **kwargs):
        def eval_wrapper(operand):
            return self.evaluate_object(operand, **kwargs)

        return [eval_wrapper(o) for o in operands]

    def evaluate(self, **kwargs):
        if not kwargs.get("cache", False):
            self.operands = self.original_operands

        self.operands = self.preprocess_operands(*self.operands, **kwargs)

        function_kw = {}

        for k in self.PASS_KWARGS:
            if k in kwargs:
                function_kw[k] = kwargs[k]

        try:
            try:
                self.rhs_index = -1
                value = self.function(*self.operands)
            except TypeError:
                value = self.operands[0]

                for i, o in enumerate(self.operands[1:]):
                    self.rhs_index = i
                    value = self.function(value, o)

            if hasattr(self.__class__, "output_cls"):
                return self.evaluate_object(value, self.output_cls, **kwargs)

            return value

        except ZeroDivisionError:
            zero = self.operands[1:].index(0) + 1
            zero_op = self.original_operands[zero]
            offset = zero_op.location - self.location
            msg = "Division by zero"

            if not isinstance(zero_op, int):
                msg += " (%s evaluated to 0)" % zero_op

            raise self.fatal(msg, offset=offset)

    @property
    def function(self):
        raise NotImplementedError("Operator subclass has no function")


class IntegerOperator(Operator):
    def preprocess_operands(self, *operands, **kwargs):
        def eval_wrapper(operand):
            return self.evaluate_object(operand, Integer, **kwargs)

        return [eval_wrapper(o) for o in operands]


class RHSIntegerOperator(IntegerOperator):
    """Like IntegerOperator, but doesn't transform the left operator to an int"""

    def preprocess_operands(self, *operands, **kwargs):
        ret = [self.evaluate_object(operands[0], **kwargs)]

        for operand in operands[1:]:
            ret.append(self.evaluate_object(operand, Integer, **kwargs))

        return ret


class Div(RHSIntegerOperator):
    
    def function(self, left, scalar):
        try:
            left = IntegerList(left)
            scalar = int(scalar)

            for i, x in enumerate(left):
                left[i] = x // scalar

            return left
        except TypeError as e:
            return Integer(perator.floordiv(left, scalar))


class Mul(RHSIntegerOperator):

    def function(self, left, scalar):
        try:
            left = IntegerList(left)
            scalar = int(scalar)

            for i, x in enumerate(left):
                left[i] = x * scalar

            return left
        except TypeError as e:
            return Integer(operator.mul(left, scalar))


class Sub(RHSIntegerOperator):
    
    def function(self, left, scalar):
        try:
            left = IntegerList(left)
            scalar = int(scalar)

            for i, x in enumerate(left):
                left[i] = x - scalar

            return left
        except TypeError as e:
            return Integer(operator.sub(left, scalar))


class Add(RHSIntegerOperator):

    def function(self, left, scalar):
        try:
            left = IntegerList(left)
            scalar = int(scalar)

            for i, x in enumerate(left):
                left[i] = x + scalar

            return left
        except TypeError as e:
            return Integer(operator.add(left, scalar))


class Modulo(RHSIntegerOperator):
    
    def function(self, left, scalar):
        try:
            left = IntegerList(left)
            scalar = int(scalar)

            for i, x in enumerate(left):
                left[i] = x % scalar

            return left
        except TypeError as e:
            return Integer(operator.mod(left, scalar))


class ComparisonOperator(Operator):

    token = ''

    def _impl(self, left, right):
        raise NotImplementedError()

    def function(self, left, right):
        if isinstance(left, IntegerList) and \
           isinstance(right, IntegerList):

            if len(left) != len(right):
                raise TypeError('Element-wise comparison requires lists of equal length')
            
            result = []
            for l, r in zip(left, right):
                value, delta = self._impl(l, r)
                result.append(Comparison(l, r, delta, value, self.token))
            return result

        elif isinstance(left, IntegerList):
            result = []
            for i in left:
                value, delta = self._impl(i, right)
                result.append(Comparison(i, right, delta, value, self.token))
            return result
        elif isinstance(right, IntegerList):
            result = []
            for i in right:
                value, delta = self._impl(left, i)
                result.append(Comparison(left, i, delta, value, self.token))
            return result
        else:
            value, delta = self._impl(left, right)
            return [Comparison(left, right, delta, value, self.token)]


class Equals(ComparisonOperator):

    token = '=='

    def _impl(self, left, right):
        result = left == right
        delta = abs(right - left)
        return result, delta


class LessThan(ComparisonOperator):

    token = '<'

    def _impl(self, left, right):
        result = left < right
        delta = abs(right - left) + (1 if not result else -1)
        return result, delta


class LessThanEqual(ComparisonOperator):

    token = '<='

    def _impl(self, left, right):
        result = left <= right
        delta = abs(right - left)
        return result, delta


class GreaterThan(ComparisonOperator):

    token = '>'

    def _impl(self, left, right):
        result = left > right
        delta = abs(right - left) + (1 if not result else -1)
        return result, delta


class GreaterThanEqual(ComparisonOperator):

    token = '>='

    def _impl(self, left, right):
        result = left >= right
        delta = abs(left - right)
        return result, delta


class AddEvenSubOdd(Operator):
    function = add_even_sub_odd


class Total(Operator):
    output_cls = Integer
    function = sum


class Successes(RHSIntegerOperator):
    def function(self, iterable, thresh):
        if not isinstance(iterable, IntegerList):
            iterable = (iterable,)
        elif isinstance(iterable, Roll):
            max_value = iterable.random_element.max_value

            if isinstance(max_value, RandomElement):
                raise self.fatal(
                    "Nested dice in success not yet supported.",
                    location=max_value.location,
                )

            if thresh > iterable.random_element.max_value:
                raise self.fatal("Success threshold higher than roll result.")

        return sum(x >= thresh for x in iterable)


class SuccessFail(RHSIntegerOperator):
    def function(self, iterable, thresh):
        result = 0
        if not isinstance(iterable, IntegerList):
            iterable = (iterable,)
        elif isinstance(iterable, Roll):
            max_value = iterable.random_element.max_value

            if isinstance(max_value, RandomElement):
                raise self.fatal(
                    "Nested dice in success not yet supported.",
                    location=max_value.location,
                )

            if thresh > iterable.random_element.max_value:
                raise self.fatal(
                    "Success threshold higher than maximum roll " "result."
                )

        if isinstance(iterable, Roll):
            fail_level = iterable.random_element.min_value
        else:
            fail_level = 1

        for x in iterable:
            if x >= thresh:
                result += 1
            elif x <= fail_level:
                result -= 1

        return result


class Again(RHSIntegerOperator):
    def function(self, lhs, rhs=None):

        if not isinstance(lhs, IntegerList):
            lhs = IntegerList([lhs])

        if rhs is None:
            if not isinstance(lhs, Roll):
                raise self.fatal("%s is not a random element" % lhs)

            rhs = lhs.random_element.max_value

        ret = lhs.copy()
        ret.clear()

        for elem in lhs:
            ret.append(elem)

            if elem == rhs:
                ret.append(elem)

        return ret


class Sort(Operator):
    def function(self, iterable):
        if not isinstance(iterable, IntegerList):
            raise self.fatal("Cannot sort %s!" % iterable)

        iterable = iterable.copy()
        iterable.sort()
        return iterable


class Extend(Operator):
    def function(self, *args):
        ret = IntegerList()
        for x in args:
            try:
                ret.extend(x)
            except TypeError:
                ret.append(x)

        return ret


class Array(Operator):
    def function(self, *args):
        print('Array')
        ret = IntegerList()

        for x in args:
            try:
                x = sum(x)
            except TypeError:
                pass
            ret.append(x)

        return ret


# TODO: stable removal instead of sort -> slice -> shuffle
class Lowest(RHSIntegerOperator):
    PASS_KWARGS = ("random",)

    def function(self, iterable, n=None, **kwargs):
        if not isinstance(iterable, IntegerList):
            raise self.fatal("Can't take the lowest values of a scalar!")

        if n is None:
            n = len(iterable) - 1

        iterable = iterable.copy()
        iterable.sort()
        iterable[n:] = []
        kwargs.get("random", random).shuffle(iterable)
        return iterable


class Highest(RHSIntegerOperator):
    PASS_KWARGS = ("random",)

    def function(self, iterable, n=None, **kwargs):
        if not isinstance(iterable, IntegerList):
            raise self.fatal("Can't take the highest values of a scalar!")

        if n is None:
            n = len(iterable) - 1

        iterable = iterable.copy()
        iterable.sort()
        iterable[:-n] = []
        kwargs.get("random", random).shuffle(iterable)
        return iterable


class Middle(RHSIntegerOperator):
    PASS_KWARGS = ("random",)

    def function(self, iterable, n=None, **kwargs):
        if not isinstance(iterable, IntegerList):
            raise self.fatal("Can't take the middle values of a scalar!")

        num = len(iterable)

        if n is None:
            n = (num - 2) if num > 2 else 1
        elif n <= 0:
            n += num

        num_remove = num - n
        upper = num_remove // 2
        lower = num_remove - upper

        iterable = iterable.copy()
        iterable.sort()
        iterable[:lower], iterable[-upper:] = [], []
        kwargs.get("random", random).shuffle(iterable)
        return iterable


class Explode(RHSIntegerOperator):
    def function(self, roll, thresh=None):
        if not isinstance(roll, Roll):
            raise self.fatal("Cannot explode {0}".format(roll))
        elif thresh is None:
            thresh = roll.random_element.max_value

        if roll.random_element.min_value == roll.random_element.max_value:
            raise self.fatal("Cannot explode a roll of one-sided dice.")

        elif thresh <= roll.random_element.min_value:
            offset = 0
            orig_thresh = self.original_operands[self.rhs_index]

            if thresh is not None:
                offset = orig_thresh.location - self.location

            msg = (
                "Refusing to explode with threshold less than or equal to "
                "the lowest possible roll."
            )

            if type(orig_thresh) is not Integer:
                msg += " (%s evaluated to %s)" % (orig_thresh, thresh)

            raise self.fatal(msg, offset=offset)

        explosions = 0
        result = list(roll)
        rerolled = roll

        while rerolled:
            explosions += 1

            if explosions >= MAX_EXPLOSIONS:
                raise self.fatal("Too many explosions!")

            num_rerolls = sum(x >= thresh for x in rerolled)
            rerolled = roll.do_roll(num_rerolls)
            result.extend(rerolled)

        return ExplodedRoll(roll.random_element, rolled=result)


class Reroll(RHSIntegerOperator):
    def function(self, roll, thresh=None):
        if not isinstance(roll, Roll):
            raise self.fatal("Cannot reroll {0}".format(roll))

        elem = roll.random_element

        if isinstance(elem.min_value, RandomElement):
            raise self.fatal(
                "Nested dice in reroll not yet supported.",
                location=elem.min_value.location,
            )

        if thresh is None:
            thresh = elem.min_value

        roll = Roll(elem, rolled=roll, force_extreme=roll.force_extreme)

        for i, x in enumerate(roll):
            if x <= thresh:
                roll[i] = roll.do_roll_single()

        return roll


class ForceReroll(RHSIntegerOperator):
    def function(self, roll, thresh=None, force_min=False):
        if not isinstance(roll, Roll):
            raise self.fatal("Cannot reroll {0}".format(roll))

        elem = roll.random_element

        if isinstance(elem.max_value, RandomElement):
            raise self.fatal(
                "Nested dice in force-reroll not yet supported.",
                location=elem.max_value.location,
            )

        if thresh is None:
            thresh = elem.min_value

        max_min = min((elem.max_value, thresh + 1))

        roll = Roll(elem, rolled=roll, force_extreme=roll.force_extreme)

        for i, x in enumerate(roll):
            if x <= thresh:
                roll[i] = roll.do_roll_single(min_value=max_min)

        return roll


class Identity(Operator):
    # no function defined because of passthrough __new__
    def __new__(self, x):
        return x


class Negate(Operator):
    def __new__(cls, x):
        if isinstance(x, int):
            # Passthrough to prevent Negate() clutter
            return Integer(-x)

        return super(Negate, cls).__new__(cls)

    def function(self, operand):
        operand = IntegerList(operand)

        for i, x in enumerate(operand):
            operand[i] = -x

        return operand




"""
Usage:
    roll [--verbose] [--min | --max] [--max-dice=<dice>] [--] <expression>...

Options:
    -m --min              Make all rolls the lowest possible result
    -M --max              Make all rolls the highest possible result
    -D --max-dice=<dice>  Set the maximum number of dice per element
    -h --help             Show this help text
    -v --verbose          Show additional output
    -V --version          Show the package version
"""

from __future__ import absolute_import, print_function, unicode_literals

import docopt

from zardoz import dice
from zardoz.dice.exceptions import DiceBaseException

__version__ = "dice v{0} by {1}".format(dice.__version__, dice.__author__)


def main(argv=None):
    """Run roll() from a command line interface"""
    args = docopt.docopt(__doc__, argv=argv, version=__version__)
    verbose = bool(args["--verbose"])

    f_roll = dice.roll
    kwargs = {}

    if args["--min"]:
        f_roll = dice.roll_min
    elif args["--max"]:
        f_roll = dice.roll_max

    if args["--max-dice"]:
        try:
            kwargs["max_dice"] = int(args["--max-dice"])
        except ValueError:
            print("Invalid value for --max-dice: '%s'" % args["--max-dice"])
            exit(1)

    expr = " ".join(args["<expression>"])

    try:
        roll, kwargs = f_roll(expr, raw=True, return_kwargs=True, **kwargs)

        if verbose:
            print("Result: ", end="")

        print(str(roll.evaluate_cached(**kwargs)))

        if verbose:
            print("Breakdown:")
            print(dice.utilities.verbose_print(roll, **kwargs))
    except DiceBaseException as e:
        print("Whoops! Something went wrong:")
        print(e.pretty_print())
        exit(1)



#!/usr/bin/env python
import curses

from sys import stdin, stdout


class raw_terminal(object):
    """ change terminal mode that the input is unbuffered """
    def __init__(self, fd):
        self.fd = fd

    def __enter__(self):
        curses.initscr()

        if not curses.termname().startswith(b'xterm'):
            raise Exception(f"This backend only supports xterm as terminal emulator. Got: '{curses.termname()}'.")

        curses.cbreak()

    def __exit__(self, type, value, traceback):
        curses.endwin()


def read_until(fd, delim):
    out = ""
    c = stdin.read(1)
    while c != delim:
        out += c
        c = stdin.read(1)
    return out


def xterm_pixels():
    """ Get the width and heigth in pixels of the xterm window """
    code = "\x1b\x5b\x31\x34\x74\x0a"

    with raw_terminal(stdin):
        stdout.write(code)
        stdout.flush()

        c = stdin.read(1)
        assert c == '\x1b'
        read_until(stdin, ';')
        height = read_until(stdin, ';')
        width = read_until(stdin, 't')

    return (int(width), int(height))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : utils.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import argparse
from dataclasses import is_dataclass
from datetime import datetime
import functools
import os
from pathlib import Path
import random

import discord
from xdg import xdg_data_home


ZARDOZ_PKG_DIR = os.path.abspath(os.path.dirname(__file__))
__time_format__ = '%a %b %d %I:%M%p %Z'

FAILURE = 'F 👎'
SUCCESS = 'S 👍'


SYNTAX = '''```
dX: Roll an X-sided die.
    Ex: d20
NdX: roll N X-sided dice.
    Ex: 3d6
NdXt: roll N x-sided dice and sum result.
    Ex: 3d6t
NdX [+,-,*,/,%] M: Apply operation with M to each roll result.
    Ex: 3d10 + 4; 3d6t + 4; 3d6 - 3; 10d2 % 2
NdX [<,<=,==,>,>=] M: Compare each roll to M. M may be integer, list, or roll.
    Ex: 3d10 < 5; 3d10 <= 3d10; 3d10 < 2d5t; 3d10 <= 4,5,6
NdX^M: Roll NdX and report the highest M results.
    Ex: 5d10^2
NdX | MdY: Concatenate roll results.
    Ex: 5d10 | 10d6
```

The roll parser is a modified form of the `python-dice` library. For more advanced usage, see https://pypi.org/project/dice/.

'''


def d10():
    return random.randint(1, 10)


def d100():
    return random.randint(1, 100)


def Nd10(n=1):
    return [random.randint(1,10) for _ in range(n)]


def Nd100(n=1):
    return [random.randint(1,100)  for _ in range(n)]


def default_database_dir(debug=False):
    if not debug:
        return xdg_data_home().joinpath('zardoz-bot', 'databases')
    else:
        return xdg_data_home().joinpath('zardoz-bot', 'debug', 'databases')


def default_log_file(debug=False):
    if not debug:
        return xdg_data_home().joinpath('zardoz-bot', 'bot.log')
    else:
        return xdg_data_home().joinpath('zardoz-bot', 'debug', 'bot.log')


def reverse_number(num: int):
    result = 0
    while (num > 0):
        result = (num * 10) + (num % 10)
        num = num // 10
    return result


def handle_http_exception(func):

    @functools.wraps(func)
    async def wrapper(self, ctx, *args, **kwargs):
        try:
            return await func(self, ctx, *args, **kwargs)
        except discord.HTTPException:
            await ctx.message.reply('Hey broh, that\'s a big response.'\
                                    ' Try dialing back your request,'\
                                    ' I\'m only human.')
            raise
    return wrapper

